#include <cstdio>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>

#include "papu.hh"
#include "clock.h"
#include "system.hh"

void bisqpapu::hz60ticks()
{
    chan1.hz60tick();
	chan2.hz60tick();
	chan3.hz60tick(),
	chan4.hz60tick();
}
void bisqpapu::hz120ticks()
{
	chan1.hz120tick();
    chan2.hz120tick();
}
void bisqpapu::hz240ticks()
{
	chan1.hz240tick();
    chan2.hz240tick();
    chan3.hz240tick();
    chan4.hz240tick();
}
void bisqpapu::hzticks()
{
	hz240ticks();
	if(--hzpos & 1)hz120ticks();
	if(!hzpos){hz60ticks();hzpos=4;}
}

bisqpapu::bisqpapu() : Rate(AudioSystem.Hz()), hzpos(4)
{
    counter = 0;
    
    // cputickrate = clockhz / cpuclockdivider
    // bytespertick = rate / cputickrate
    // bytespertick = rate * cpuclockdivider / clockhz
    
    speed = (unsigned long)(Clock->CPUclockDivider * Rate * (double)(1 << PAPU_FIXEDPOINT) / Clock->ClockHz);
}

bisqpapu::~bisqpapu()
{
}

bool bisqpapu::write(const fWORD addr, fBYTE value)
{
    switch(addr)
    {
#if PAPU_CH1_ENABLED
        case 0x4000 ... 0x4003: fprintf(stderr, "\33[1m1\33[m");chan1.write(addr&3, value); return true;
#endif
#if PAPU_CH2_ENABLED
        case 0x4004 ... 0x4007: fprintf(stderr, "\33[1m2\33[m");chan2.write(addr&3, value); return true;
#endif
#if PAPU_CH3_ENABLED
        case 0x4008 ... 0x400B: fprintf(stderr, "\33[1m3\33[m");chan3.write(addr&3, value); return true;
#endif
#if PAPU_CH4_ENABLED
        case 0x400C ... 0x400F: fprintf(stderr, "\33[1m4\33[m");chan4.write(addr&3, value); return true;
#endif
#if PAPU_CH5_ENABLED
        case 0x4010 ... 0x4013: fprintf(stderr, "\33[1m5\33[m");chan5.write(addr&3, value); return true;
#endif
        case 0x4015:
            if(value & 0x01)chan1.enable();else chan1.disable();
            if(value & 0x02)chan2.enable();else chan2.disable();
            if(value & 0x04)chan3.enable();else chan3.disable();
            if(value & 0x08)chan4.enable();else chan4.disable();
            if(value & 0x10)chan5.enable();else chan5.disable();
            // &32, &64 and &128 ignored.
            return true;
    }
    return false;
}
bool bisqpapu::read(const fWORD addr, BYTE &value)
{
    switch(addr)
    {
        case 0x4000 ... 0x4013:
            // Complete. These are write-only.
            value = 0x40;
            return true;
        case 0x4015:
        //  fprintf(stderr, "\33[1;32mpAPU status: ");
        	
        	/* These bits really do mean the status of the length unit */
            if(!chan1.lengthunit_silent())value |= 0x01;
            if(!chan2.lengthunit_silent())value |= 0x02;
            if(!chan3.lengthunit_silent())value |= 0x04;
            if(!chan4.lengthunit_silent())value |= 0x08;
            
            if(!chan5.silent())value |= 0x10;
            if(chan5.irqflag)value |= 0x80;
        //  fprintf(stderr, "%02Xn\33[m\n", value);
            return true;
    }
    return false;
}

void bisqpapu::mixbytes(unsigned count)
{
    static const BYTE triangledata[32]=
    {13,14,15,15, 14,13,12,11, 10, 9, 8, 7,  6, 5, 4, 3,
      2, 1, 0, 0,  1, 2, 3, 4,  5, 6, 7, 8,  9,10,11,12};
    
    // chan1 = +-15, chan2 = +-15
    // chan3 = 
    
    int val1 = (chan1.silent() || (chan1.dutycounter&15) < chan1.dutytype) ? 0 : chan1.getvol();
    int val2 = (chan2.silent() || (chan2.dutycounter&15) < chan2.dutytype) ? 0 : chan2.getvol();
    int val3 = chan3.silent() ? 0 : triangledata[chan3.trianglecounter&31]; // No getvol() here
    int val4 = chan4.silent() ? 0 : chan4.output * chan4.getvol();
    
    // Those were four 4-bit DACs.
    
    int val5 = chan5.deltacounter;
    
    // Channel 1,2,4 outputs are BIT (0 OR 1) times VOLUME  - 0..15
    // Channel 3 is TRIANGLE                                - 0..15
    // Channel 5 is DIGITAL                                 - 0..127
    
    int pin1 = val1 + val2;        // 0..30
    int pin2 = val3 + val4;        // 0..30
    
    SWORD val = (((pin1 + pin2) // range: 0..60
                       << 1)          // range: 0..120
                       - val5)        // range: -127..120
                       * (32768/128);
    AudioSystem.Write(val, count);
}

bool bisqpapu::tick()
{
    // This function runs Clock->ClockHz / Clock->CPUclockDivider
    // times per second. For NTSC, it is 1.7897725 MHz.

	if(!--hzcounter)
	{
		hzcounter = countermax;
		hzticks();
	}

#if PAPU_CH1_ENABLED
    chan1.tick();
#endif
#if PAPU_CH2_ENABLED
    chan2.tick();
#endif
#if PAPU_CH3_ENABLED
    chan3.tick();
#endif
#if PAPU_CH4_ENABLED
    chan4.tick();
#endif
#if PAPU_CH5_ENABLED
    bool irq = chan5.tick();
#else
	bool irq = false;
#endif
    counter += speed;
    if(counter >= (1 << PAPU_FIXEDPOINT))
    {
        mixbytes(counter >> PAPU_FIXEDPOINT);
        counter &= (1 << PAPU_FIXEDPOINT)-1;
    }

    return irq;
}
